package org.bitbucket.openisoj;

import java.util.HashMap;

import org.bitbucket.openisoj.exceptions.UnknownFieldException;

public abstract class AMessage {
	public class AccountType {
		public static final String _00_DEFAULT = "00";
		public static final String _10_SAVINGS = "10";
		public static final String _20_CHECK = "20";
		public static final String _30_CREDIT = "30";
		public static final String _40_UNIVERSAL = "40";
		public static final String _50_INVESTMENT = "50";
		public static final String _60_ELECTRONIC_PURSE_DEFAULT = "60";
	}

	public class AmountType {
		public static final String _01_LEDGER_BALANCE = "01";
		public static final String _02_AVAILABLE_BALANCE = "02";
		public static final String _03_OWING = "03";
		public static final String _04_DUE = "04";
		public static final String _20_REMAINING_THIS_CYCLE = "20";
		public static final String _40_CASH = "40";
		public static final String _41_GOODS_SERVICES = "41";
		public static final String _53_APPROVED = "53";
		public static final String _56_TIP = "56";
		public static final String _90_AVAILABLE_CREDIT = "90";
		public static final String _91_CREDIT_LIMIT = "91";

	}

	private Bitmap _bitmap;

	private HashMap<Integer, IField> _fields;

	protected Template template;

	public AMessage(Template template) {
		this.template = template;
		_fields = new HashMap<Integer, IField>();
		_bitmap = new Bitmap();
	}

	public void clearField(int field) {
		_bitmap.setField(field, false);
		_fields.remove(field);
	}

	protected abstract void constructTemplate() throws Exception;

	protected abstract IField createField(int field) throws UnknownFieldException;

	protected String describePacking() throws Exception {
		return template.describePacking();
	}

	public String get(int field) {
		return getFieldValue(field);
	}

	protected IField getField(int field) throws UnknownFieldException {
		if (!_bitmap.isFieldSet(field) || !_fields.containsKey(field)) {
			_fields.put(field, createField(field));
			_bitmap.setField(field, true);
		}

		return _fields.get(field);
	}

	public String getFieldValue(int field) {
		return _bitmap.isFieldSet(field) ? _fields.get(field).getValue() : null;
	}

	public int getPackedLength() throws Exception {
		return getBasePackedLength();
	}

	private int getBasePackedLength() throws Exception {
		int length = _bitmap.getPackedLength();
		for (int i = 2; i <= 128; i++)
			if (_bitmap.isFieldSet(i))
				length += _fields.get(i).getPackedLength();

		return length;
	}

	public ProcessingCode getProcessingCode() throws Exception {
		return new ProcessingCode(_fields.get(3).getValue());
	}

	public boolean isFieldSet(int field) {
		return _bitmap.isFieldSet(field);
	}

	public void set(int field, String value) throws UnknownFieldException {
		setFieldValue(field, value);
	}

	protected void setFieldValue(int field, String value) throws UnknownFieldException {
		if (value == null) {
			clearField(field);
			return;
		}

		getField(field).setValue(value);
	}

	protected byte[] toMsg() throws Exception {
		int packedLen = getBasePackedLength();
		byte[] data = new byte[packedLen];

		int offset = 0;

		byte[] bmap = _bitmap.toMsg();
		System.arraycopy(bmap, 0, data, offset, _bitmap.getPackedLength());
		offset += _bitmap.getPackedLength();

		for (int i = 2; i <= 128; i++) {
			if (_bitmap.isFieldSet(i)) {
				IField field = _fields.get(i);
				System.arraycopy(field.toMsg(), 0, data, offset, field.getPackedLength());
				offset += field.getPackedLength();
			}
		}

		return data;
	}

	@Override
	public String toString() {
		try {
			return toString("");
		} catch (Exception e) {
			e.printStackTrace();
			return "ERROR GENERATING STRING";
		}
	}

	protected String toString(int field, String prefix) throws Exception {
		return _fields.get(field).toString(prefix + "   ");
	}

	protected String toString(String prefix) throws Exception {
		StringBuilder sb = new StringBuilder();

		for (int i = 2; i <= 128; i++)
			if (_bitmap.isFieldSet(i))
				sb.append(toString(i, prefix));

		return sb.toString();
	}

	protected int unpack(byte[] msg, int startingOffset) throws Exception {
		int offset = startingOffset;

		offset = _bitmap.unpack(msg, offset);

		for (int i = 2; i <= 128; i++)
			if (_bitmap.isFieldSet(i))
				offset = getField(i).unpack(msg, offset);

		return offset;
	}
}
