package org.bitbucket.openisoj;

import org.bitbucket.openisoj.formatter.BinaryFormatter;
import org.bitbucket.openisoj.formatter.IFormatter;

public class Bitmap {
	private IFormatter _formatter;
	private boolean[] _bits;

	public Bitmap() {
		this(new BinaryFormatter());
	}

	public Bitmap(IFormatter formatter) {
		_formatter = formatter;
		_bits = new boolean[128];
	}

	public int getPackedLength() throws Exception {
		return _formatter.getPackedLength(isExtendedBitmap() ? 32 : 16);
	}

	public boolean isExtendedBitmap() {
		return isFieldSet(1);
	}

	public boolean isFieldSet(int field) {
		return _bits[field - 1];
	}

	public void setField(int field, boolean on) {
		_bits[field - 1] = on;
		_bits[0] = false;

		for (int i = 64; i < 127; i++)
			if (_bits[i]) {
				_bits[0] = true;
				break;
			}
	}

	public byte[] toMsg() throws Exception {
		int lengthOfBitmap = isExtendedBitmap() ? 16 : 8;

		byte[] data = new byte[lengthOfBitmap];

		for (int i = 0; i < lengthOfBitmap; i++)
			for (int j = 0; j < 8; j++)
				if (_bits[i * 8 + j])
					data[i] = (byte) (data[i] | (128 / (int) Math.pow(2, j)));

		if (_formatter instanceof BinaryFormatter)
			return data;

		IFormatter binaryFormatter = new BinaryFormatter();
		String bitMapString = binaryFormatter.getString(data);

		return _formatter.getBytes(bitMapString);
	}

	public int unpack(byte[] msg, int offset) throws Exception {
		int lengthOfBitmap = _formatter.getPackedLength(16);

		if (_formatter instanceof BinaryFormatter) {
			// Java needs the 0xFF because otherwise it get's confused
			// wtf explained http://stackoverflow.com/questions/792386
			if ((msg[offset] & 0xFF) >= 128) {
				lengthOfBitmap += 8;
			}
		} else if (msg[offset] >= 0x38) {
			lengthOfBitmap += 16;
		}

		byte[] bitmapData = new byte[lengthOfBitmap];
		System.arraycopy(msg, offset, bitmapData, 0, lengthOfBitmap);

		if (!(_formatter instanceof BinaryFormatter)) {
			IFormatter binaryFormatter = new BinaryFormatter();
			String value = _formatter.getString(bitmapData);
			bitmapData = binaryFormatter.getBytes(value);
		}

		for (int i = 0; i < bitmapData.length; i++)
			for (int j = 0; j < 8; j++)
				_bits[i * 8 + j] = (bitmapData[i] & (128 / (int) Math.pow(2, j))) > 0;

		return offset + lengthOfBitmap;
	}
}
